;;########################################################################
;; visuals.lsp 
;; copyright (c) 1991-95 by Forrest W. Young
;; Contains code for the ViSta version of Visuals to create a guided tour 
;; object proto and a function to construct a guided tour spreadplot.
;;########################################################################

(require "vista")

(defmeth mv-data-object-proto :visualize-tour-plot ()
  (let* ((vars 
          (column-list (send current-data :active-data-matrix '(numeric))))
         (num-unique 
          (length (remove-duplicates (round (mapcar #'variance vars)))))
         (pt-labs (send current-data :active-labels))
         (var-labs (send current-data :active-variables '(numeric))))
    (if (= 1 num-unique)
        (visuals vars pt-labs var-labs nil :scale-type nil)
        (error "You must equate variances and means of all variables before using the Guided Tour."))))

;;************************************************************************
;; define guided tour prototype based on visuals guided tour algorithm
;;************************************************************************

(defproto tour-proto 
  '(visuals-tour min-degrees deg rok radians incremental-rotation model-p) 
   () spin-proto)

(defmeth tour-proto :visuals-tour (&optional (arg nil set))
"Determines type of tour algorithm:  t for visuals tour, nil for screen tour."
  (if set (setf (slot-value 'visuals-tour) arg))
  (slot-value 'visuals-tour))

(send tour-proto :visuals-tour t)

(defmeth tour-proto :min-degrees (&optional (arg nil set))
"Minimum rotation increment in degrees for tour algorithm."
  (if set (setf (slot-value 'min-degrees) arg))
  (slot-value 'min-degrees))

(send tour-proto :min-degrees -2.8125) ;;minimum hD-spin increment in degrees
  
(defmeth tour-proto :deg (&optional (arg nil set))
"Current rotation increment of tour algorithm on each of three axes, in degrees."
  (if set (setf (slot-value 'deg) arg))
  (slot-value 'deg))

(send tour-proto :deg (repeat (send tour-proto :min-degrees) 3))

(defmeth tour-proto :rok (&optional (arg nil set))
"Used to reverse rotation direction for rocking capability."
  (if set (setf (slot-value 'rok) arg))
  (slot-value 'rok))

(defmeth tour-proto :radians (&optional (arg nil set))
"Current rotation increment of tour algorithm on each of three axes, in radians."
  (if set (setf (slot-value 'radians) arg))
  (slot-value 'radians))

(defun radians (deg) (/ (* deg pi) 360))

(send tour-proto :radians (mapcar #'radians (send tour-proto :deg)))

(defmeth tour-proto :incremental-rotation (&optional (arg nil set))
"Current incremental rotation matrix.  Specifies one rotation step of the tour algorithm."
  (if set (setf (slot-value 'incremental-rotation) arg))
  (slot-value 'incremental-rotation))

(defun six-dim-rotate (a b c ndim)
  (let ((i-mat (identity-matrix ndim)))
    (matmult (make-rotation (row i-mat 0) (row i-mat 3) (radians a))
             (make-rotation (row i-mat 1) (row i-mat 4) (radians b))
             (make-rotation (row i-mat 2) (row i-mat 5) (radians c))))) 

(defmeth tour-proto :model-p (&optional (arg nil set))
"Model-p is t if model being visualized, nil if data being visualized."
  (if set (setf (slot-value 'model-p) arg))
  (slot-value 'model-p))

;;------------------------------------------------------------------------
;;define touring methods for tour-proto (to be activated by buttons)
;;------------------------------------------------------------------------

(send tour-proto :scale-type 'fixed :draw nil)

(defmeth tour-proto :adjust-to-data ()
"The adjust-to-data method for tour-plots centers the plot a the origin."
  (send self :center (iseq (send self :num-variables))  0))
 
(defmeth tour-proto :do-idle ()
"Method for guided tour to perform one step in the tour while system is 
otherwise idle."
  (let ((trans (send self :transformation))
        (radians (send self :radians))
        )
    (cond 
      ((not (send self :visuals-tour))
       (send self :rotate-2 0 3 (select radians 0) :draw nil)
       (send self :rotate-2 1 4 (select radians 1)))
      (
;This is the original Visuals method.  It doesn't work right and the
;menu item to activate it is commented out of the code below.
       (if trans (send self :transformation 
                       (matmult trans (send self :incremental-rotation)))
                 (send self :transformation 
                       (send self :incremental-rotation)))))))

(defmeth tour-proto :tour-alg ()
"Method for guided tour to toggle between the two available tour algorithms by toggeling value of visuals-tour.  When visuals-tour is true, the algorithm is the original visuals algorithm.  When visuals-tour is nil,  the algorithm is a new algorithm based on screen horizontal and vertical that is faster than the visuals algorithm.  Both are equivalent until 3D rotations (the pitch, roll and yaw buttons) are used."
  (send self :visuals-tour (not (send self :visuals-tour)))
  (send self :redraw)
  )

(defmeth tour-proto :tour-speed (axis dir)
"Method for guided tour to control tour speed on each tour dimension."
  (let* ((min-degrees (send self :min-degrees))
         (deg (send self :deg))
         (degnow (select deg axis))
         (rok (send self :rok))
         )
    (if (= degnow 0)
        (if dir 
            (if rok (setf (select deg axis) min-degrees)
                (setf (select deg axis) (* -1 min-degrees)))))
    (if (/= degnow 0)
        (if dir (setf (select deg axis) (* (select deg axis) 2)) 
            (setf (select deg axis) (/ (select deg axis) 2))))
    (if (> (select deg axis) 360) (setf (select deg axis) 360))
    (if (<  min-degrees (select deg axis) (- min-degrees))
        (setf (select deg axis) 0))
    (send self :incremental-rotation 
          (six-dim-rotate (select deg 0) (select deg 1) (select deg 2)
                          (send self :num-variables)))
    (send self :radians (mapcar #'radians deg))
    (send self :deg deg)
  ))

(defmeth tour-proto :rock ()
"Method to rock the tour plot by reversing rotation direction."
  (let ((deg (send self :deg (- (send self :deg))))
        )
    (send self :rok (not (send self :rok)))
    (send self :radians (mapcar #'radians deg))
    (send self :incremental-rotation 
          (six-dim-rotate (select deg 0) (select deg 1) (select deg 2)
                          (send self :num-variables)))))

(defmeth tour-proto :home ()
"Method to return the tour plot to its home orientation.  The method also stops rotation and resets rotation increments to their default values and directions."
  (let* ((min-degrees (send self :min-degrees))
         (deg (send self :deg))
         (point-state (send self :point-state (iseq (send self :num-points))))
         )
    (send self :idle-on nil)
    (send self :rok     nil)
    (send self :deg     (list min-degrees min-degrees min-degrees))
    (send self :radians (mapcar #'radians (send self :deg)))
    (send self :transformation nil)
    (send target1 :transformation nil)
    (send target2 :transformation nil)
    (when (send self :model-p) 
          (send ingram1 :transformation nil)
          (send ingram2 :transformation nil))
    (send tour-plot :point-state 
          (iseq (send tour-plot :num-points)) point-state)
    ))

(defmeth tour-proto :stop-go ()
"Method to stop or start guided tour."
  (let ((npts (send self :num-points))
        )
    (send self :idle-on (not (send self :idle-on)))
    (send self :point-state (iseq npts) (send self :point-state (iseq npts)))
    ))

(defmeth tour-proto :new-tour ()
"Method for guided tour to perform residualization."
  (let* ((nobs (send self :num-points))
         (nvar (send self :num-variables))
         (nvar-1 (- nvar 1))
         (transf (send self :transformation))
         (point-state (send self :point-state (iseq nobs)))
         (svd-struct nil)
         )
    (when transf
          (setf svd-struct 
               (sv-decomp (select transf (iseq nvar) (iseq 3 nvar-1))))
          (setf transf 
               (bind-rows 
                (select transf (iseq 3) (iseq nvar)) 
                (transpose (matmult (select svd-struct 0)
                                    (diagonal (select svd-struct 1))))))
          (send tour-plot :transformation transf)
          
          (send target1 :transformation transf)
          (send target2 :transformation transf)
          (when (send self :model-p) 
                (send ingram1 :transformation transf)
                (send ingram2 :transformation transf)))
    (send tour-plot :point-state (iseq nobs) point-state)
    ))

(defmeth tour-proto :isnew (&rest args)
  (apply #'call-next-method args)
  (send self :add-overlay (send tour-overlay :new)))

(defmeth tour-proto :make-menu-item (item)
  (if (symbolp item)
      (case item
        (tour-method (send menu-item-proto :new "Tour Method"
            :action #'(lambda () (send self :tour-alg))))
        (t (call-next-method item)))
      item))

(send tour-proto :menu-template (append (send spin-proto :menu-template)
                                        '(tour-method)))

;;------------------------------------------------------------------------
;;define overlay to contain buttons for tour-proto
;;------------------------------------------------------------------------

(defproto tour-overlay () () graph-overlay-proto)

(let* ((ascent (send tour-proto :text-ascent))
       (x ascent)
       (y (round (* 1.5 ascent)))
       (box ascent))
  (send tour-overlay :add-slot 'location
        (list x y box (round (+ x (* 1.5 box))))))

(defmeth tour-overlay :location () (slot-value 'location))

;;------------------------------------------------------------------------
;;define buttons to overlay onto hD Spin-Plot
;;------------------------------------------------------------------------

(defmeth tour-overlay :redraw ()
  (let* ((loc (send self :location))
         (x (first loc))
         (y (second loc))
         (box (third loc))
         (string-x (fourth loc))
         (graph (send self :graph))
         (ch (send tour-proto :canvas-height))
         (visuals-tour (send graph :visuals-tour))
         (button-strings nil)
         (k nil)
         (l nil)
         (xoffsets '(0 68 135 202))
         (yoffsets '(0 18))
         (soffsets '(0 15)))

    (if visuals-tour
        (setf button-strings 
             '("Go/Stop" "Home" "Rock" "New Tour" "X Spin" "Y Spin" "Z Spin"))
        (setf button-strings 
             '("Go/Stop" "Home" "Rock" "New Tour" "H Spin" "V Spin" "")))

    (setf k  0)
    (setf l -1)
    (dolist (j yoffsets)
      (setf l (+ l 1))
       (dolist (i xoffsets)
         (when (not (and (= i 202) (= j 18)))
               (when (not (and (= i 135) (= j 18) (not visuals-tour)))
                     (send graph :frame-rect (+ i x) (+ j (- y box)) box box)
                     (if (= j 18)
                         (send graph :frame-rect 
                               (+ i 15 x) (+ j (- y box)) box box))
                     (if (/= j 18)
                         (setf soff 0) 
                         (setf soff (select soffsets l)))
                     (send graph :draw-string (select button-strings k) 
                           (+ i soff string-x) (+ j y))
)
               (setf k (+ k 1)))))
    (send graph :frame-rect 220 (- ch 2 box) box box)
    (send graph :frame-rect 235 (- ch 2 box) box box)
    (send graph :draw-line 0 37 (send graph :canvas-width) 37)
    ))

;;------------------------------------------------------------------------
;;connect buttons to tour-proto methods
;;------------------------------------------------------------------------

(defmeth tour-overlay :do-click (x y m1 m2)
  (let* ((loc (send self :location))
         (box (third loc))
         (left (first loc))
         (top (- (second loc) box))          
         (right (+ left box))
         (bottom (+ top box))
         (ch (send tour-proto :canvas-height))
         (graph (send self :graph))
         (pt-st nil)
         )
    (setf pt-st (send graph :point-state (iseq (send graph :num-points))))
    (cond 
      ((and (< left x right) (< (+ 18 top) y (+ 18 bottom)))
       (send graph :tour-speed 0 nil))
      ((and (< (+  15 left) x (+  15 right)) (< (+ 18 top) y (+ 18 bottom)))
       (send graph :tour-speed 0 t))
      ((and (< (+  68 left) x (+  68 right)) (< (+ 18 top) y (+ 18 bottom)))
       (send graph :tour-speed 1 nil))
      ((and (< (+  83 left) x (+  83 right)) (< (+ 18 top) y (+ 18 bottom)))
       (send graph :tour-speed 1 t))
      ((and (< (+ 135 left) x (+ 135 right)) (< (+ 18 top) y (+ 18 bottom)))
       (send graph :tour-speed 2 nil))
      ((and (< (+ 150 left) x (+ 150 right)) (< (+ 18 top) y (+ 18 bottom)))
       (send graph :tour-speed 2 t))
      ((and (< left x right) (< top y bottom))
       (send graph :stop-go))  
      ((and (< (+ 68 left) x (+ 68 right)) (< top y bottom))
       (send graph :home))
      ((and (< (+ 135 left) x (+ 135 right)) (< top y bottom))
       (send graph :rock))
      ((and (< (+ 202 left) x (+ 202 right)) (< top y bottom))
       (send graph :new-tour)))
    ))


;;########################################################################
;;define guided tour constructor function
;;########################################################################

(defun tour-plot (data &rest args)
  (let* ((graph (apply #'send tour-proto :new (length data) args))
        )
    (send graph :add-points data)
    (send graph :margin 0 38 0 18)
    (send graph :incremental-rotation
          (let ((deg (send graph :deg))
                )  
            (six-dim-rotate (select deg 0) (select deg 1) (select deg 2)
                            (send graph :num-variables))
            ))
    graph))



;;########################################################################
;;define visuals guided tour spreadplot plot-cells
;;########################################################################

(defun visuals (vars point-labels var-labels model-p &key scale-type)
"Args: (variables point-labels variables-labels model-p &key scale-type)
When more than six variables, displays a guided tour consisting of a scatterplot-matrix, high-dimensional spin-plot, and two 3d spin-plots.  Returns an error message when less than six variables.  When more than nine variables uses only first nine.  Each plot is formed from VARIABLES, and uses POINT-LABLES and VARIABLE-LABELS.  Model-p is t when creating a guided tour of a model, nil for guided tours of data (no model). Scale-type can be nil, fixed or variable.  These work as described for other plots."
(let ((nvar (length vars))
      (variables vars)
      (variable-labels var-labels))
  (when (> nvar 9)
        (setf nvar 9)
        (setf variables (select variables (iseq 9)))
        (setf variable-labels (select variable-labels (iseq 9)))
        (error-message "Too many variables for Guided Tour. The first 9 are used."))
  (if (< nvar 6)
      (too-few)
      (tour-spreadplot variables point-labels variable-labels
                       model-p scale-type nvar))))

(defun too-few ()
  (error-message (format nil 
     "Too few variables for Guided Tour.  There must be at least 6.")))

(defun tour-spreadplot 
  (variables point-labels variable-labels model-p scale-type nvar)
"Args: (variables point-labels variables-labels &key scale-type)
Used by the visuals function to display the visuals guided tour spreadplot."
  (let ((permuted-variables 
         (select variables (combine '(0 1 2 3 4) (iseq 5 (- nvar 1)))))
        (permuted-variable-labels
         (select variable-labels (combine '(0 1 2 3 4) (iseq 5 (- nvar 1)))))
        )

  (setf scatmat (scatterplot-matrix variables
                :show nil
                :size (floor (*  1.2 graph-size))
                :location 
                  (list 1  (- 
                   (+ (select (- screen-size (floor (*  1.2 graph-size))) 1) 
                      window-decoration-height) msdos-fiddle))
                :scale-type scale-type
                :point-labels point-labels
                :variable-labels variable-labels))

  (send scatmat  :add-mouse-mode 'focus-on-variables
        :title "Focus On Variables"
        :click :do-variable-focus
        :cursor 'finger)
  (send scatmat  :mouse-mode 'focus-on-variables)
  (defmeth scatmat :do-variable-focus (x y m1 m2)
    (do-variable-focus-visuals x y m1 m2 model-p))
  
  (setf tour-plot (tour-plot permuted-variables
                :show nil
               
                :size (list (select (floor (*  1.2 graph-size)) 0)
                         (+ (select (floor (*  1.2 graph-size)) 1) 54))
                :location (list (+ (select (floor (* .8 graph-size)) 0)
                                   window-decoration-width)
                                (select location12 1))
                :scale-type scale-type
                :title "Guided Tour"
                :point-labels point-labels
                :variable-labels permuted-variable-labels))

  (send tour-plot :point-label 
        (iseq (send tour-plot :num-points))
        point-labels)
        
  (send tour-plot :scale (iseq 9) .4)
  (send tour-plot :model-p model-p)

  (send tour-plot :new-menu "TourPlot" 
        :items '(SHOWING-LABELS MOUSE RESIZE-BRUSH SLICER DASH
                 ERASE-SELECTION FOCUS-ON-SELECTION SHOW-ALL SYMBOL 
                 COLOR SELECTION DASH FASTER SLOWER CUING AXES REVERSE
 ;               TOUR-METHOD commented out because visuals method not right
                 ))

  (send tour-plot :tour-alg)
;  (send tour-plot :show-window)

  (setf  target1 (spin-plot permuted-variables
                :show nil
                :size (floor (* .8 graph-size))
                :location location11
                :scale-type scale-type
                :title "Left Target"
                :point-labels point-labels
                :variable-labels permuted-variable-labels))

  (send target1 :new-menu "Target1" 
        :items '(SHOWING-LABELS MOUSE RESIZE-BRUSH SLICER DASH
                 ERASE-SELECTION FOCUS-ON-SELECTION SHOW-ALL SYMBOL 
                 COLOR SELECTION DASH FASTER SLOWER CUING AXES REVERSE))
  
  (defmeth target1 :adjust-to-data () 
    (send self :center (iseq (send self :num-variables))  0))

  (send target1 :content-variables 0 1 2)
  (send target1 :scale (iseq 9) .4)
  (send target1 :adjust-to-data)
;  (send target1 :show-window)
  
  (setf  target2 (spin-plot permuted-variables
                :show nil
                
                :size (floor (* .8 graph-size))
                :location (list (- (select location13 0) 3)
                                (select location13 1))
                :scale-type scale-type
                :title "Right Target"
                :point-labels point-labels
                :variable-labels permuted-variable-labels))

  (send target2 :new-menu "Target2" 
        :items '(SHOWING-LABELS MOUSE RESIZE-BRUSH SLICER DASH
                 ERASE-SELECTION FOCUS-ON-SELECTION SHOW-ALL SYMBOL 
                 COLOR SELECTION DASH FASTER SLOWER CUING AXES REVERSE))


  (defmeth target2 :adjust-to-data () 
    (send self :center (iseq (send self :num-variables))  0))

  (send target2 :content-variables 3 4 5)
  (send target2 :scale (iseq 9) .4)

    (setf obs-list 
         (name-list (send current-data :labels) 
              :location (list 
               (- (select location13 0) 3)
               (- (+ (select (- screen-size (floor (*  1.2 graph-size))) 1) 
                     window-decoration-height) msdos-fiddle))))
    (send obs-list :size 
          (floor (* .8 (+ window-decoration-width (select graph-size 0))))
          (floor (select (* 1.2 graph-size) 1)))

  (send scatmat   :linked t)
  (send tour-plot :linked t)
  (send target1   :linked t)
  (send target2   :linked t)
  (send obs-list  :linked t)

    (when model-p  
        (send current-model :create-ingram-plots
              permuted-variables point-labels
                    permuted-variable-labels))

    (defmeth target1 :close () 
      (send self :remove)
      (send tour-plot :remove)
      (send target2 :remove)
      (send scatmat :remove)
      (send obs-list :remove))

    (defmeth tour-plot :close ()
      (send target1 :close))
    
    (defmeth target2 :close ()
      (send target1 :close))
    
    (defmeth scatmat :close ()
      (send target1 :close))

    (defmeth obs-list :close ()
      (send target1 :close))
    (send target1 :show-window)
    (send target2 :show-window)
    (send obs-list :show-window)
    (send scatmat :show-window)
    (send tour-plot :show-window)
  ))

(defun do-variable-focus-visuals (x y m1 m2 model-p)
"Args: (x y m1 m2 model-p)
Function used by the scatterplot matrix do-variable-focus method for the visuals guided tour to focus on variable subplots. X and Y specify which subplot is selected, m1 and m2 specify whether shift or option keys are down (when t), model-p specifies if model visualization (when t). Assumes that three other plots exist which are named tour-plot, target1 and target2.  Clicks on sub-plots reset the selection list to nil.  Shift-clicks send first two selected variables to target1 and tour-plot as first two variables, second two selected variables to target2  and tour-plot (as variables 4 and 5) and third two select variables to all three plots as last variables."
  (let* ((nvar (send scatmat :num-variables))
         (nobs (send scatmat :num-points))
         (cur-var (send scatmat :current-variables))
         (variable-labels (send scatmat :variable-label (iseq nvar)))
        )
    (when (not m1)
          (setf tour-var ())
          )
    (when m1 ;when shift-click
          (when (= (select cur-var 0) (select cur-var 1))
                (setf cur-var (list (select cur-var 0))))
          (when (< (length tour-var) 6)
                (setf tour-var (adjoin (select cur-var 0) tour-var))
                (when (< (length tour-var) 6)
                      (if (equal (length cur-var) 2)
                          (setf tour-var 
                                (adjoin (select cur-var 1) tour-var)))))
          (when (= (length tour-var) 6)
                (setf tour-var (reverse tour-var))
                (if (> nvar 6)
                    (setf tour-var (combine tour-var 
                         (reverse (set-difference (iseq nvar) tour-var)))))
                (new-home nvar nobs 
                 (select tour-var (combine '(0 1 4 2 3) (iseq 5 (- nvar 1))))
                 variable-labels model-p)
                (setf tour-var ())
                (send tour-plot :show-window)
                ))
    ))

(defun new-home (nvar nobs tour-var variable-labels model-p)
"Args: (nvar nobs tour-var variable-labels)
Used by scattterplot matrix do-variable-focus method with visuals spreadplot to create a new home set of variables for the grand tour.  Nvar and nobs are number of variables and observations in the tour. Tour-var is a list of the new home variables. Variable-labels is a list of the variable labels."
  (dotimes (i nvar)
           (send tour-plot :point-coordinate i (iseq nobs)
                 (send scatmat :point-coordinate 
                       (select tour-var i) (iseq nobs)))
           (send target1   :point-coordinate i (iseq nobs)
                 (send scatmat :point-coordinate 
                       (select tour-var i) (iseq nobs)))
           (send target2   :point-coordinate i (iseq nobs)
                 (send scatmat :point-coordinate 
                       (select tour-var i) (iseq nobs))))
  (send tour-plot :variable-label  (iseq nvar)
        (select variable-labels tour-var))
  (send target1   :variable-label (iseq nvar) 
        (select variable-labels tour-var))
  (send target2   :variable-label (iseq nvar) 
        (select variable-labels tour-var))
  (if model-p (send current-model :update-ingram-plots
                    nvar nobs tour-var variable-labels))
  (send tour-plot :home)
  )

(setf tour-var ())